#define RAY
//#define SCREEN_SPACE

// gameObjectExtensions.cs 
// Written by Vegard Myklebust at useful slug
// This script enables you to do 3d interface coding in a simple way
// as an extension of the GameObject class.

using UnityEngine;
using System.Collections;
using System.Collections.Generic;

public static class GameObjectExtensions
{
	#region VISUAL STATE CHANGES - EDIT FOR YOUR APPLICATION
	private static bool default_state_change = true; // Change to false if you want your default .Clicked() etc calls to not change visual state
	private static void MouseOverState(GameObject click_object)
	{
		// TODO Change whatever you want about the state here, this is an example
		if (click_object.animation != null)
			click_object.animation.CrossFade("MouseOver");
		for (int i=0;i<click_object.renderer.materials.Length;i++)
		{
			if (click_object.renderer.materials[i].name == "Light (Instance)"){
				click_object.renderer.materials[i].SetFloat("_Emission",0.3f);
			}
		}
	}
	
	private static void MouseDownState(GameObject click_object)
	{
		// TODO Change whatever you want about the state here, this is an example
		for (int i=0;i<click_object.renderer.materials.Length;i++)
		{
			if (click_object.renderer.materials[i].name == "Light (Instance)"){
				click_object.renderer.materials[i].SetFloat("_Emission",1.0f);
			}
		}
	}
	
	private static void MouseOutState(GameObject click_object)
	{
		// TODO Change whatever you want about the state here, this is an example
		if (click_object.animation != null)
			click_object.animation.CrossFade("MouseOut");
		for (int i=0;i<click_object.renderer.materials.Length;i++)
		{
			if (click_object.renderer.materials[i].name == "Light (Instance)"){
				click_object.renderer.materials[i].SetFloat("_Emission",0.0f);
			}
		}
	}
	#endregion VISUAL STATE CHANGES
	
	#region EXTENSION METHODS 
	private static bool button_down;
	private static GameObject active_object;
	private static List<GameObject> check_box_objects = new List<GameObject>();
	private static List<GameObject> radio_object_group = new List<GameObject>();
	
	public static bool Clicked(this GameObject click_object, int mouse_button, Camera camera, bool change_state)
	{
		bool return_value;
		if (Input.GetMouseButtonDown(mouse_button))
		{
			return_value = Hover(click_object, camera, false);
			if (change_state && return_value)
				MouseDownState(click_object);
			return(return_value);
		}
		else if (Input.GetMouseButton(mouse_button)){
			return_value = Hover(click_object, camera, false);
			if (change_state && return_value)
				MouseDownState(click_object);
			return(false);
		}
		else
			return(false);
	}
	
	public static bool CheckBox(this GameObject click_object, int mouse_button, Camera camera, bool change_state)
	{
		if (Input.GetMouseButtonDown(mouse_button) || (Input.multiTouchEnabled && Input.touchCount > 0))
		{
			if (Hover(click_object, camera, false))
			{
				if (check_box_objects.Contains(click_object)) {
					check_box_objects.Remove(click_object);
				}
				else {
					check_box_objects.Add(click_object);
				}
			}
			
		}
		if (check_box_objects.Contains(click_object)) {
			if (change_state)
				MouseDownState(click_object);
			return(true);
		}
		else {
			if (change_state){
				if (Hover(click_object, camera, false))
					MouseOverState(click_object);
				else
					MouseOutState(click_object);
			}
			return(false);
		}
	}
	
	public static bool RadioButton(this GameObject click_object, int radio_group_id, int mouse_button, Camera camera, bool change_state)
	{
		// We just maintain a single list of all active radio buttons, all others are false
		while (radio_object_group.Count < radio_group_id+1) {
			radio_object_group.Add(click_object);
		}
		if (Input.GetMouseButtonDown(mouse_button) || (Input.multiTouchEnabled && Input.touchCount > 0))
		{
			if (Hover(click_object, camera, false))
			{
				radio_object_group.Insert(radio_group_id,click_object);
			}
		}
		if (radio_object_group[radio_group_id] == click_object)
		{
			if (change_state)
				MouseDownState(click_object);
			return(true);
		}
		else
		{
			if (change_state)
				MouseOutState(click_object);
			return(false);
		}
	}
	
	public static void RestoreState(this GameObject click_object)
	{
		MouseOutState(click_object);
	}
	
	public static bool ClickHold(this GameObject click_object, int mouse_button, Camera camera, bool change_state)
	{
		bool return_value;
		if (Input.GetMouseButton(mouse_button))
		{
			return_value = Hover(click_object, camera, false);
			if (change_state && return_value)
				MouseDownState(click_object);
			return(return_value);
		}
		else if (Input.multiTouchEnabled && Input.touchCount > 0)
		{
			return_value = Hover(click_object, camera, false);
			if (change_state && return_value)
				MouseDownState(click_object);
			return(return_value);
		}
		else
			return(false);
	}
	
	public static bool Hover(this GameObject click_object, Camera camera, bool change_state)
	{
		Bounds bounds = click_object.renderer.bounds;
		#if RAY
			Ray screen_ray;
			Vector3 input_position;
			if (Input.multiTouchEnabled)
			{
				if (Input.touchCount > 0)
				{
					for (int i = 0;i<Input.touchCount;i++)
					{
						input_position = new Vector3(Input.touches[i].position.x,Input.touches[i].position.y,0);
						screen_ray = camera.ScreenPointToRay(input_position);
						if (bounds.IntersectRay(screen_ray))
						{
							if(change_state)
								MouseOverState(click_object);
							return(true);
						}
					}
					if(change_state)
						MouseOutState(click_object);
					return(false);
				}
				else
				{	
					if(change_state)
						MouseOutState(click_object);
					return(false);
				}
			}
			else
			{
				input_position = Input.mousePosition;
				screen_ray = camera.ScreenPointToRay(input_position);
				if (bounds.IntersectRay(screen_ray))
				{
					if(change_state)
						MouseOverState(click_object);
					return(true);
				}
				else
				{
					if(change_state)
						MouseOutState(click_object);
					return(false);
				}
			}
		
		#elif SCREEN_SPACE
			Vector4 bounding_square = GetScreenBoundingSquare(bounds, camera);
			float x = Input.mousePosition.x;
			float y = Input.mousePosition.y;
			if (x > bounding_square[0] && x < bounding_square[2] && y > bounding_square[1] && y < bounding_square[3])
			{
				return(true);
			}
			else
				return(false);
		#endif	
	}
	
	public static bool Drag(this GameObject click_object, int mouse_button, Camera camera, bool change_state)
	{
		Vector3 mouse_delta = DragDelta(click_object, mouse_button, change_state);
		click_object.transform.Translate(mouse_delta);
		if (button_down == true && active_object == click_object)
			return(true);
		else
			return(false);
	}
	
	public static float DragSlider(this GameObject click_object, int mouse_button, int axis, float min, float max, Vector3 start, Vector3 end, Camera camera, bool change_state)
	{
		float return_value; 
		Vector3 mouse_delta;
		mouse_delta = DragDelta(click_object, mouse_button, camera, change_state);	
		Vector3 direction = end - start;
		Vector3 offset = direction.normalized * mouse_delta[axis];
		
		click_object.transform.Translate(offset);
		
		// Get t, the percentage of the slider
		float t = ValueOnLineSegment(start, end, click_object.transform.localPosition);
		if (t <= 0)
			click_object.transform.localPosition = start;
		else if (t >= 1)
			click_object.transform.localPosition = end;
		return_value = min + (max-min)*t;
		return(return_value);
	}
	
	public static Vector3 DragDelta(this GameObject click_object, int mouse_button, Camera camera, bool change_state)
	{
		Vector3 return_vector = new Vector3(0,0,0);
		if (Input.GetMouseButtonDown(mouse_button))
		{
			if (Hover(click_object, camera, false))
			{
				button_down = true;
				active_object = click_object;
			}
		}
		if (button_down && active_object == click_object)
		{	
			Ray screen_ray = camera.ScreenPointToRay(Input.mousePosition);
			float distance = Vector3.Distance(screen_ray.origin,click_object.transform.position);
			return_vector = (screen_ray.origin + screen_ray.direction.normalized * distance) - click_object.transform.position;
			if (change_state)
				MouseDownState(click_object);
		}
		if (Input.GetMouseButtonUp(mouse_button))
		{
			button_down = false;
			if (change_state)
				MouseOutState(click_object);
		}		
		
		return(return_vector);
	}
	#endregion EXTENSION METHODS
	
	#region OVERLOADS
	// CLICKED
	public static bool Clicked(this GameObject click_object)
	{
		Camera camera = Camera.main;
		return(Clicked(click_object, 0, camera, default_state_change));
	}
	
	public static bool Clicked(this GameObject click_object, bool change_state)
	{
		Camera camera = Camera.main;
		return(Clicked(click_object, 0, camera, change_state));
	}
	
	public static bool Clicked(this GameObject click_object, int mouse_button)
	{
		Camera camera = Camera.main;
		return(Clicked(click_object, mouse_button, camera, default_state_change));
	}
	
	public static bool Clicked(this GameObject click_object, int mouse_button, bool change_state)
	{
		Camera camera = Camera.main;
		return(Clicked(click_object, mouse_button, camera, change_state));
	}
	
	public static bool Clicked(this GameObject click_object, int mouse_button, Camera camera)
	{
		return(Clicked(click_object, mouse_button, camera, default_state_change));
	}
	// CHECKBOX
	public static bool CheckBox(this GameObject click_object)
	{
		Camera camera = Camera.main;
		return(CheckBox(click_object, 0, camera, default_state_change));
	}
	
	public static bool CheckBox(this GameObject click_object, bool change_state)
	{
		Camera camera = Camera.main;
		return(CheckBox(click_object, 0, camera, change_state));
	}	
	
	public static bool CheckBox(this GameObject click_object, int mouse_button, bool change_state)
	{
		Camera camera = Camera.main;
		return(CheckBox(click_object, mouse_button, camera, change_state));
	}
	
	public static bool CheckBox(this GameObject click_object, int mouse_button, Camera camera)
	{
		return(CheckBox(click_object, mouse_button, camera, default_state_change));
	}
	// HELD
	public static bool ClickHold(this GameObject click_object)
	{
		Camera camera = Camera.main;
		return(ClickHold(click_object, 0, camera, default_state_change));
	}
	
	public static bool ClickHold(this GameObject click_object, bool change_state)
	{
		Camera camera = Camera.main;
		return(ClickHold(click_object, 0, camera, change_state));
	}
	
	public static bool ClickHold(this GameObject click_object, int mouse_button)
	{
		Camera camera = Camera.main;
		return(ClickHold(click_object, mouse_button, camera, default_state_change));
	}
	
	public static bool ClickHold(this GameObject click_object, int mouse_button, Camera camera)
	{
		return(ClickHold(click_object, mouse_button, camera, default_state_change));
	}
	
	public static bool ClickHold(this GameObject click_object, int mouse_button, bool change_state)
	{
		Camera camera = Camera.main;
		return(ClickHold(click_object, mouse_button, camera, change_state));
	}
	// HOVER
	public static bool Hover(this GameObject click_object)
	{
		Camera camera = Camera.main;
		return(Hover(click_object,camera,default_state_change));	
	}
	
	public static bool Hover(this GameObject click_object, bool change_state)
	{
		Camera camera = Camera.main;
		return(Hover(click_object,camera,change_state));	
	}
	
	public static bool Hover(this GameObject click_object, Camera camera)
	{
		return(Hover(click_object,camera,default_state_change));	
	}
	// DELTADRAG
	public static Vector3 DragDelta(this GameObject click_object)
	{
		Camera camera = Camera.main;
		return(DragDelta(click_object,0,camera,default_state_change));
	}
	
	public static Vector3 DragDelta(this GameObject click_object, bool change_state)
	{
		Camera camera = Camera.main;
		return(DragDelta(click_object,0,camera,change_state));
	}
	
	public static Vector3 DragDelta(this GameObject click_object, int mouse_button)
	{
		Camera camera = Camera.main;
		return(DragDelta(click_object,mouse_button,camera,default_state_change));
	}
	
	public static Vector3 DragDelta(this GameObject click_object, int mouse_button, bool change_state)
	{
		Camera camera = Camera.main;
		return(DragDelta(click_object,mouse_button,camera,change_state));
	}
	
	public static Vector3 DragDelta(this GameObject click_object, int mouse_button, Camera camera)
	{
		return(DragDelta(click_object,mouse_button,camera,default_state_change));
	}
	// DRAG
	public static bool Drag(this GameObject click_object)
	{
		Camera camera = Camera.main;
		return(Drag(click_object,0,camera, default_state_change));
	}
	
	public static bool Drag(this GameObject click_object, bool change_state)
	{
		Camera camera = Camera.main;
		return(Drag(click_object,0,camera, change_state));
	}
	
	public static bool Drag(this GameObject click_object, int mouse_button)
	{
		Camera camera = Camera.main;
		return(Drag(click_object,mouse_button,camera, default_state_change));
	}
	
	public static bool Drag(this GameObject click_object, int mouse_button, bool change_state)
	{
		Camera camera = Camera.main;
		return(Drag(click_object,mouse_button,camera, change_state));
	}
	
	public static bool Drag(this GameObject click_object, int mouse_button, Camera camera)
	{
		return(Drag(click_object,mouse_button,camera, default_state_change));
	}
	// DRAGSLIDER
	public static float DragSlider(this GameObject click_object, int mouse_button, int axis, float min, float max, Vector3 start, Vector3 end)
	{
		Camera camera = Camera.main;
		return(DragSlider(click_object, mouse_button, axis, min, max, start, end, camera, default_state_change));
	}
	
	public static float DragSlider(this GameObject click_object, int mouse_button, int axis, float min, float max, Vector3 start, Vector3 end, bool change_state)
	{
		Camera camera = Camera.main;
		return(DragSlider(click_object, mouse_button, axis, min, max, start, end, camera, change_state));
	}
	
	public static float DragSlider(this GameObject click_object, int mouse_button, int axis, float min, float max, Vector3 start, Vector3 end, Camera camera)
	{
		return(DragSlider(click_object, mouse_button, axis, min, max, start, end, camera, default_state_change));
	}
	// RADIOBUTTON
	public static bool RadioButton(this GameObject click_object, int radio_group_id)
	{
		Camera camera = Camera.main;
		return(RadioButton(click_object, radio_group_id, 0, camera, default_state_change));
	}
	
	public static bool RadioButton(this GameObject click_object, int radio_group_id, bool change_state)
	{
		Camera camera = Camera.main;
		return(RadioButton(click_object, radio_group_id, 0, camera, change_state));
	}
	
	public static bool RadioButton(this GameObject click_object, int radio_group_id, int mouse_button)
	{
		Camera camera = Camera.main;
		return(RadioButton(click_object, radio_group_id, mouse_button, camera, default_state_change));
	}
	
	public static bool RadioButton(this GameObject click_object, int radio_group_id, int mouse_button, bool change_state)
	{
		Camera camera = Camera.main;
		return(RadioButton(click_object, radio_group_id, mouse_button, camera, change_state));
	}
	
	public static bool RadioButton(this GameObject click_object, int radio_group_id, int mouse_button, Camera camera)
	{
		return(RadioButton(click_object, radio_group_id, mouse_button, camera, default_state_change));
	}
	#endregion OVERLOADS
	
	#region Helper functions
	
	private static float ValueOnLineSegment(Vector3 A, Vector3 B, Vector3 P)
	{
	    Vector3 AP = P - A;
	    Vector3 AB = B - A;
	    float ab2 = AB.x*AB.x + AB.y*AB.y + AB.z*AB.z;
	    float ap_ab = AP.x*AB.x + AP.y*AB.y + AP.z*AB.z;
	    float t = ap_ab / ab2;
	    
		if (t < 0.0f) t = 0.0f;
	    else if (t > 1.0f) t = 1.0f;
		
		return(t);
	}
	
	private static double RayPlaneIntersection(Ray ray, Vector3 point_on_plane, Vector3 normal)
	{
		double distance = Vector3.Dot(point_on_plane-ray.origin,normal) / Vector3.Dot(ray.direction,normal);
		return(distance);
	}
	
		
	private static Vector4 GetScreenBoundingSquare(Bounds bounds, Camera camera)
	{
		Vector3 screen_vector;
		Vector3[] box_vectors = new Vector3[8];
		Vector4 return_vector; // We return a single vector4 with minimum in the first two and max in the second two
		
		// We put the box vertices in an array for easy looping, not strictly necessary
		box_vectors[0] = new Vector3(bounds.min[0],bounds.min[1],bounds.min[2]); // Vertex 0 all minimum
		box_vectors[1] = new Vector3(bounds.min[0],bounds.min[1],bounds.max[2]); // Vertex 1
		box_vectors[2] = new Vector3(bounds.min[0],bounds.max[1],bounds.max[2]); // Vertex 2
		box_vectors[3] = new Vector3(bounds.min[0],bounds.max[1],bounds.min[2]); // Vertex 3
		box_vectors[4] = new Vector3(bounds.max[0],bounds.max[1],bounds.max[2]); // Vertex 4 all maximum, the other side of the box
		box_vectors[5] = new Vector3(bounds.max[0],bounds.max[1],bounds.min[2]); // Vertex 5
		box_vectors[6] = new Vector3(bounds.max[0],bounds.min[1],bounds.min[2]); // Vertex 6
		box_vectors[7] = new Vector3(bounds.max[0],bounds.min[1],bounds.max[2]); // Vertex 7
		
		// Initialize
		screen_vector = camera.WorldToScreenPoint(box_vectors[0]);
		return_vector = new Vector4(screen_vector[0],screen_vector[1],screen_vector[0],screen_vector[1]);
		
		for (int i=0;i<8;i++)
		{
			screen_vector = camera.WorldToScreenPoint(box_vectors[i]);
			if (screen_vector[0] < return_vector[0])
				return_vector[0] = screen_vector[0];
			if (screen_vector[1] < return_vector[1])
				return_vector[1] = screen_vector[1];
			if (screen_vector[0] > return_vector[2])
				return_vector[2] = screen_vector[0];
			if (screen_vector[1] > return_vector[3])
				return_vector[3] = screen_vector[1];
		}
		
		return(return_vector);
	}
	
	#endregion
}
